import datetime

from ..commander import Command, Commander


class ClickCommander(Commander):
    """Generates commands based on the command config"""

    def add_base_imports(self):
        self.cli = f"""## Generated {self.manifest.name} on {datetime.datetime.now()}
import rich_click as click
import subprocess
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
"""

    def add_base_cli(self):
        self.cli += f"""
@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option('{self.manifest.version}')
def cli():
    pass
"""

    def add_group_command(self, command: Command):
        self.cli += f"""
@cli.command()
def {command.name}():
    \"\"\"Help for {command.name}\"\"\"
    {self.parser.parse_command(command.script)}
"""

    def add_sub_command(self, command: Command, group: str):
        self.cli += f"""
@{group}.command()
def {command.name}():
    \"\"\"Help for {command.name}\"\"\"
    {self.parser.parse_command(command.script)}
"""



from collections import defaultdict
from typing import DefaultDict

from pybash.transformer import transform as transform_bash
from pydantic import BaseModel

from .manifests import COMMAND_BLOCK, Manifest
from .parser import Parser


class Command(BaseModel):
    name: str
    script: COMMAND_BLOCK


class CLI(BaseModel):
    name: str
    version: str
    code: str
    requires: list[str] = []


class Group(BaseModel):
    name: str
    commands: list[Command]
    help: str = ""


class Commander:
    """Generates commands based on the command config"""

    __slots__ = ("manifest", "parser", "cli", "groups", "greedy", "commands", "root_commands")

    def __init__(self, manifest: Manifest) -> None:
        self.manifest = manifest
        self.parser = Parser(self.manifest)
        self.cli: str = ""
        self.groups: dict[str, Group] = {}
        self.greedy: list[Command] = []
        self.commands: list[Command] = [
            Command(name=name, script=script) for name, script in self.manifest.commands.items()
        ]
        self.root_commands: list[Command] = [command for command in self.commands if "." not in command.name]
        self.build_groups()

    def build_groups(self) -> None:
        groups: DefaultDict[str, list[Command]] = defaultdict(list)
        group_help_dict: dict[str, str] = {}
        for command in self.commands:
            # Check for greedy commands- evaluate them at the end
            if self.is_greedy(command.name):
                self.greedy.append(command)
                continue

            if "." in command.name:
                group_name = command.name.split(".")[:-1][-1]
                groups[group_name].append(command)
            else:
                for script_block in command.script:
                    if isinstance(script_block, dict) and script_block.get("help"):
                        group_help = script_block["help"]
                        group_help_dict[command.name] = group_help

        for group_name, commands in groups.items():
            self.groups[group_name] = Group(
                name=group_name, commands=commands, help=group_help_dict.get(group_name, "")
            )

    def build_cli(self) -> None:
        self.add_base_imports()
        self.add_imports()
        self.add_vars()
        self.add_base_cli()
        self.add_functions()
        self.add_root_commands()
        self.add_subcommands()
        self.add_greedy_commands()
        self.add_main_block()

    def add_root_commands(self) -> None:
        for root_command in self.root_commands:
            self.add_root_command(root_command)

    def add_subcommands(self) -> None:
        for group in self.groups.values():
            self.add_group(group)
            for subcommand in group.commands:
                self.add_sub_command(subcommand, group)

    def add_imports(self) -> None:
        if not self.manifest.imports:
            return

        if isinstance(self.manifest.imports, str):
            self.cli += self.manifest.imports + "\n"
        elif isinstance(self.manifest.imports, list):
            for _import in self.manifest.imports:
                self.cli += _import + "\n"
        self.cli += "\n"

    def add_vars(self) -> None:
        if not self.manifest.vars:
            return

        for var, val in self.manifest.vars.items():
            if isinstance(val, dict):
                self.cli += f"{var} = {next(iter(val))}\n"
            else:
                self.cli += f"{var} = '{val}'\n"
        self.cli += "\n"

    def add_functions(self) -> None:
        if not self.manifest.functions:
            return

        for func in self.manifest.functions:
            self.cli += f"{transform_bash(func)}\n"
        self.cli += "\n"

    def add_command(self, command: Command) -> None:
        if "." in command.name:
            group = command.name.split(".")[:-1][-1]
            self.add_sub_command(command, self.groups[group])

    def add_greedy_commands(self) -> None:
        """Greedy commands get lazy-loaded. Only supported for group-commands currently"""
        for greedy_command in self.greedy:
            if greedy_command.name.startswith("(*)"):
                for group in self.groups:
                    # make it lazy and interpolate
                    lazy_command_name = greedy_command.name.replace("(*)", group)
                    lazy_command_script = ""
                    if isinstance(greedy_command.script, str):
                        lazy_command_script = greedy_command.script.replace("{(*)}", group)
                    elif isinstance(greedy_command.script, list):
                        lazy_command_script = []
                        for script_block in greedy_command.script:
                            if isinstance(script_block, dict):
                                lazy_command_script.append(script_block["help"].replace("{(*)}", group))
                            else:
                                lazy_command_script.append(script_block.replace("{(*)}", group))

                    if greedy_command_args := self.manifest.args.get(greedy_command.name):
                        self.manifest.args[lazy_command_name] = greedy_command_args

                    # lazy load
                    lazy_command = Command(name=lazy_command_name, script=lazy_command_script)
                    self.commands.append(lazy_command)
                    self.add_command(lazy_command)

    def is_greedy(self, val: str) -> bool:
        """Greedy strings must contain (*)- marked to be evaluated lazily."""
        return "(*)" in val

    def add_group(self, group: Group) -> None:
        raise NotImplementedError

    def add_base_imports(self) -> None:
        raise NotImplementedError

    def add_base_cli(self) -> None:
        raise NotImplementedError

    def add_root_command(self, command: Command) -> None:
        raise NotImplementedError

    def add_group_command(self, command: Command) -> None:
        raise NotImplementedError

    def add_sub_command(self, command: Command, group: Group) -> None:
        raise NotImplementedError

    def add_main_block(self) -> None:
        raise NotImplementedError


def build_cli(manifest: Manifest, commander_cls=Commander) -> CLI:
    commander = commander_cls(manifest)
    commander.build_cli()
    return CLI(name=manifest.name, version=manifest.version, code=commander.cli, requires=manifest.requires)



## Mimic rich API methods for rich-less support
from typing import Any

import click


class Console:
    def __init__(self) -> None:
        pass

    def print(self, text: Any, **kwargs) -> None:
        if isinstance(text, Table):
            click.echo(text)
        else:
            print(text)


class Table:

    __slots__ = ("cols", "rows", "styles")

    def __init__(self) -> None:
        self.cols: list[str] = []
        self.rows: list[list[str]] = []
        self.styles: list[str] = []

    def add_column(self, col: str, style: str = "") -> None:
        self.cols.append(col)
        self.styles.append(style)

    def add_row(self, *row) -> None:
        self.rows.append([*row])

    def __str__(self) -> str:
        text = "".join([click.style(f"{col:10}", fg=self.styles.pop(0)) for col in self.cols]) + "\n"
        for row in self.rows:
            for col in row:
                text += f"{col:10}"
            text += "\n"
        return text



import datetime

from ..commander import Command, Commander, Group


class TyperCommander(Commander):
    """Generates commands based on the command config"""

    def add_base_imports(self):
        self.cli = f"""## Generated {self.manifest.name} on {datetime.datetime.now()}
import typer
import subprocess
from typing import Optional
"""

    def add_base_cli(self) -> None:
        self.cli += """
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
cli = typer.Typer(context_settings=CONTEXT_SETTINGS"""

        if self.manifest.cli_options:
            self.cli += f",{self.parser.to_args(self.manifest.cli_options)}"
        if self.manifest.help:
            self.cli += f', help="""{self.manifest.help}"""'

        self.cli += f""")
__version__ = '{self.manifest.version}'
__cli_name__ = '{self.manifest.name}'


def version_callback(value: bool):
    if value:
        print(f"{{__cli_name__}}, {{__version__}}")
        raise typer.Exit()


@cli.callback()
def main(version: Optional[bool] = typer.Option(None, '--version', callback=version_callback, is_eager=True)):
    pass

"""

    def add_root_command(self, command: Command) -> None:
        self.cli += f"""
@cli.command("{self.parser.get_parsed_command_name(command)}")
def {self.parser.get_command_func_name(command)}({self.parser.parse_args(command)}):
{self.parser.parse_command(command.script)}
"""

    def add_group(self, group: Group) -> None:
        self.cli += f"""{group.name}_app = typer.Typer()
cli.add_typer({group.name}_app, name="{group.name}", help="{group.help}")
"""

    def add_sub_command(self, command: Command, group: Group) -> None:
        self.cli += f"""
@{group.name}_app.command("{self.parser.get_parsed_command_name(command)}")
def {self.parser.get_command_func_name(command)}({self.parser.parse_args(command)}):
{self.parser.parse_command(command.script)}
"""

    def add_main_block(self) -> None:
        self.cli += """
if __name__ == "__main__":
    cli()
"""



## CLI to generate CLIs
import contextlib
from tempfile import NamedTemporaryFile
from typing import TextIO

try:
    import rich_click as click  # type: ignore
    from rich.console import Console  # type: ignore
    from rich_click.rich_group import RichGroup as AliasGroup  # type: ignore
except ImportError:
    import click
    from .rich import Console
    from click import Group as AliasGroup

from .builder import build_cli, run_cli
from .helper import CLIFFY_CLI_DIR, age_datetime, exit_err, indent_block, out, out_err, print_rich_table, write_to_file
from .homer import get_clis, get_metadata, get_metadata_path, remove_metadata, save_metadata
from .loader import Loader
from .manifests import Manifest, set_manifest_version
from .transformer import Transformer

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


class AliasedGroup(AliasGroup):  # type: ignore
    def get_command(self, ctx, cmd_name):
        with contextlib.suppress(KeyError):
            cmd_name = ALIASES[cmd_name].name
        return super().get_command(ctx, cmd_name or "")


@click.group(context_settings=CONTEXT_SETTINGS, cls=AliasedGroup)
@click.version_option()
def cli() -> None:
    pass


@cli.command()
@click.argument("manifests", type=click.File("rb"), nargs=-1)
def load(manifests: list[TextIO]) -> None:
    """Load CLI for given manifest(s)"""
    for manifest in manifests:
        T = Transformer(manifest)
        Loader.load_from_cli(T.cli)
        save_metadata(manifest.name, T.cli)
        out(f"✨ Generated {T.cli.name} CLI v{T.cli.version} ✨", fg="green")
        out("$", fg="magenta", nl=False)
        out(f" {T.cli.name} -h")


@cli.command()
@click.argument("cli_names", type=str, nargs=-1)
def update(cli_names: list[str]) -> None:
    """Reloads CLI by name"""
    for cli_name in cli_names:
        if cli_metadata := get_metadata(cli_name):
            T = Transformer(open(cli_metadata.runner_path, "r"))
            Loader.load_from_cli(T.cli)
            save_metadata(cli_metadata.runner_path, T.cli)
            out(f"✨ Reloaded {T.cli.name} CLI v{T.cli.version} ✨", fg="green")
            out("$", fg="magenta", nl=False)
            out(f" {T.cli.name} -h")
        else:
            out_err(f"~ {cli_name} not found")


@cli.command()
@click.argument("manifest", type=click.File("rb"))
def render(manifest: TextIO) -> None:
    """Render the CLI manifest generation as code"""
    T = Transformer(manifest)
    console = Console()
    console.print(T.cli.code, overflow="fold", emoji=False, markup=False)
    out(f"# Rendered {T.cli.name} CLI v{T.cli.version} ~", fg="green")


@cli.command("run")
@click.argument("manifest", type=click.File("rb"))
@click.argument("cli_args", type=str, nargs=-1)
def cliffy_run(manifest: TextIO, cli_args: tuple[str]) -> None:
    """Run CLI for a manifest"""
    T = Transformer(manifest)
    run_cli(T.cli.name, T.cli.code, cli_args)


@cli.command()
@click.argument("cli_name", type=str, default="cliffy")
@click.option("--version", "-v", type=str, show_default=True, default="v1", help="Manifest version")
@click.option("--render", is_flag=True, show_default=True, default=False, help="Render template to terminal directly")
@click.option(
    "--raw",
    type=bool,
    is_flag=True,
    show_default=True,
    default=False,
    help="Raw template without boilerplate helpers and examples.",
)
def init(cli_name: str, version: str, render: bool, raw: bool) -> None:
    """Generate a CLI manifest template"""
    set_manifest_version(version)
    template = Manifest.get_raw_template(cli_name) if raw else Manifest.get_template(cli_name)

    if render:
        console = Console()
        console.print(template, overflow="fold", emoji=False, markup=False)
    else:
        try:
            write_to_file(f"{cli_name}.yaml", text=template)
        except Exception as e:
            exit_err(f"~ error writing to file: {e}")
        out(f"+ {cli_name}.yaml", fg="green")


@cli.command("list")
def cliffy_list() -> None:
    "List all CLIs loaded"
    cols = ["Name", "Version", "Age", "Manifest"]
    rows = [
        [metadata.cli_name, metadata.version, age_datetime(metadata.loaded), metadata.runner_path]
        for metadata in get_clis()
    ]
    print_rich_table(cols, rows, styles=["cyan", "magenta", "green", "blue"])


@cli.command()
@click.argument("cli_names", type=str, nargs=-1)
def remove(cli_names: list[str]) -> None:
    "Remove a loaded CLI by name"
    for cli_name in cli_names:
        if get_metadata_path(cli_name):
            remove_metadata(cli_name)
            Loader.unload_cli(cli_name)
            out(f"~ {cli_name} removed 💥", fg="green")
        else:
            out_err(f"~ {cli_name} not loaded")


@cli.command()
@click.argument("cli_names", type=str, nargs=-1)
@click.option("--debug", is_flag=True, show_default=True, default=False, help="Display build output")
@click.option("--output-dir", "-o", type=click.Path(file_okay=False, dir_okay=True, writable=True), help="Output dir")
def bundle(cli_names: list[str], debug: bool, output_dir: str) -> None:
    "Bundle a loaded CLI into a zipapp"
    for cli_name in cli_names:
        if not (metadata := get_metadata(cli_name)):
            out_err(f"~ {cli_name} not loaded")
            continue

        result = build_cli(
            cli_name, script_path=f"{CLIFFY_CLI_DIR}/{cli_name}.py", deps=metadata.requires, output_dir=output_dir
        )

        if result.exit_code != 0:
            out(result.stdout)
            out_err(f"~ {cli_name} bundle failed")
            continue

        if debug:
            out(result.stdout)
        out(f"+ {cli_name} bundled 📦", fg="green")


@cli.command()
@click.argument("manifests", type=click.File("rb"), nargs=-1)
@click.option("--debug", is_flag=True, show_default=True, default=False, help="Display build output")
@click.option("--output-dir", "-o", type=click.Path(file_okay=False, dir_okay=True, writable=True), help="Output dir")
def build(manifests: list[TextIO], debug: bool, output_dir: str) -> None:
    "Build a CLI manifest into a zipapp"
    for manifest in manifests:
        T = Transformer(manifest, validate_requires=False)
        with NamedTemporaryFile(mode="w", prefix=f"{T.cli.name}_", suffix=".py", delete=True) as script:
            script.write(T.cli.code)
            script.flush()
            result = build_cli(T.cli.name, script_path=script.name, deps=T.cli.requires, output_dir=output_dir)

        if result.exit_code != 0:
            out(result.stdout)
            out_err(f"~ {T.cli.name} build failed")
            continue

        if debug:
            out(result.stdout)
        out(f"+ {T.cli.name} built 📦", fg="green")


@cli.command()
@click.argument("cli_name", type=str)
def info(cli_name: str):
    "Display CLI info"
    metadata = get_metadata(cli_name) or exit_err(f"~ {cli_name} not loaded")
    out(f"{click.style('name:', fg='blue')} {metadata.cli_name}")
    out(f"{click.style('version:', fg='blue')} {metadata.version}")
    out(f"{click.style('requires:', fg='blue')} {metadata.requires}")
    out(f"{click.style('age:', fg='blue')} {age_datetime(metadata.loaded)} ({metadata.loaded.ctime()})")
    out(f"{click.style('manifest:', fg='blue')}\n{indent_block(metadata.manifest, spaces=2)}")


ALIASES = {
    "add": load,
    "rm": remove,
    "ls": cliffy_list,
}



from . import v1

Manifest = v1.CLIManifest
IncludeManifest = v1.IncludeManifest
COMMAND_BLOCK = v1.COMMAND_BLOCK


def set_manifest_version(manifestVersion: str) -> None:
    global Manifest
    global IncludeManifest
    global COMMAND_BLOCK
    if manifestVersion == "v1":
        Manifest = v1.CLIManifest
        IncludeManifest = v1.IncludeManifest
        COMMAND_BLOCK = v1.COMMAND_BLOCK



import glob
import json
import os
from base64 import b32encode
from datetime import datetime
from pathlib import Path
from typing import Iterator, Optional

from .commander import CLI
from .helper import CLIFFY_HOME_PATH, write_to_file
from .manifests.v1 import CLIMetadata


def save_metadata(manifest_path: str, cli: CLI) -> None:
    """Stores CLI metadata

    Args:
        manifest_path (str): CLI manifest path
        cli (CLI): CLI
    """
    abs_manifest_path = os.path.realpath(manifest_path)
    encoded_runnerpath = b32encode(cli.name.encode("ascii")).decode("utf-8")
    save_metadata_path = f"{CLIFFY_HOME_PATH}/{encoded_runnerpath}/{cli.name}.json"
    with open(manifest_path, "r") as manifest:
        write_to_file(
            save_metadata_path,
            json.dumps(
                CLIMetadata(
                    cli_name=cli.name,
                    runner_path=abs_manifest_path,
                    version=cli.version,
                    loaded=datetime.now(),
                    manifest=manifest.read(),
                    requires=cli.requires,
                ).model_dump(),
                default=str,
            ),
        )


def remove_metadata(cli_name: str) -> None:
    """Clears CLI metadata by name

    Args:
        cli_name (str): CLI name
    """
    metadata_paths = glob.glob(f"{CLIFFY_HOME_PATH}/*/{cli_name}.*")
    for path in metadata_paths:
        parent_path = Path(path).parent.resolve()
        os.remove(path)
        os.rmdir(parent_path)


def get_metadata_bypath(path: Path) -> CLIMetadata:
    """Fetches CLI metadata by metadata path

    Args:
        path (Path): Metadata path

    Returns:
        CLIMetadata: CLI metadata
    """
    try:
        return CLIMetadata(**json.load(path.open("r")))
    except Exception as e:
        return CLIMetadata(
            cli_name=path.name,
            runner_path=path.absolute().name,
            version="error",
            loaded=datetime.now(),
            manifest=f"could not load: {e}",
            requires=[],
        )


def get_metadata(cli_name: str) -> Optional[CLIMetadata]:
    """Fetches CLI metadata by name

    Args:
        cli_name (str): CLI name

    Returns:
        Optional[CLIMetadata]: CLI metadata
    """
    cli_path = get_metadata_path(cli_name)
    if not cli_path:
        return None

    path = Path(cli_path)
    return get_metadata_bypath(path)


def get_metadata_path(cli_name: str) -> Optional[str]:
    """Fetches CLI metadata path

    Args:
        cli_name (str): CLI name

    Returns:
        Optional[str]: CLI metadata path
    """
    if cli_metadata_path := glob.glob(f"{CLIFFY_HOME_PATH}/*/{cli_name}.json"):
        return cli_metadata_path[0]
    return None


def get_clis() -> Iterator[CLIMetadata]:
    """Fetches loaded CLIs metadata iteratively

    Yields:
        Iterator[CLIMetadata]: CLI metadata
    """
    metadata_paths = Path(CLIFFY_HOME_PATH).glob("*/*")
    for metadata_path in metadata_paths:
        yield get_metadata_bypath(metadata_path)



import os
from typing import Any, TextIO

import yaml
from jinja2 import BaseLoader, Environment, FileSystemLoader
from pydantic import ValidationError
from typing_extensions import Self

from .commander import build_cli
from .commanders.typer import TyperCommander
from .helper import compare_versions, exit_err, get_installed_pip_packages, out, parse_requirement
from .manifests import IncludeManifest, Manifest, set_manifest_version
from .merger import cliffy_merger


class Transformer:
    """Loads command manifest and transforms it into a CLI"""

    __slots__ = ("manifest_io", "command_config", "manifest_version", "includes_config", "manifest", "cli")

    def __init__(self, manifest_io: TextIO, *, as_include: bool = False, validate_requires: bool = True) -> None:
        self.manifest_io = manifest_io
        self.command_config = self.load_manifest(manifest_io)
        self.manifest_version = self.command_config.pop("manifestVersion", "v1")
        if self.command_config.get("includes"):
            self.includes_config = self.resolve_includes()
            cliffy_merger.merge(self.command_config, self.includes_config)

        set_manifest_version(self.manifest_version)
        if as_include:
            try:
                self.manifest = IncludeManifest(**self.command_config)
            except ValidationError as e:
                out(f"{e}")
                exit_err(f"~ error validating {manifest_io.name}")
        else:
            try:
                self.manifest = Manifest(**self.command_config)
            except ValidationError as e:
                out(f"{e}")
                exit_err(f"~ error validating {manifest_io.name}")

            if validate_requires:
                self.validate_cli_requires()
            self.cli = build_cli(self.manifest, commander_cls=TyperCommander)

    def validate_cli_requires(self) -> None:
        if not self.manifest.requires:
            return

        installed_pip_packages = get_installed_pip_packages()
        for dep in self.manifest.requires:
            dep_spec = parse_requirement(dep)
            if dep_spec.name not in installed_pip_packages:
                exit_err(f"~ missing requirement: `{self.manifest_io.name}` requires `{dep}` to be installed")

            if (
                dep_spec.version
                and dep_spec.operator
                and not compare_versions(
                    installed_pip_packages[dep_spec.name],
                    dep_spec.version,
                    dep_spec.operator,
                )
            ):
                exit_err(
                    f"~ missing requirement: `{self.manifest_io.name}` requires `{dep}` to be installed"
                    f"    found version `{installed_pip_packages[dep_spec.name]}`"
                )

    def resolve_includes(self) -> dict:
        include_transforms = map(self.resolve_include_by_path, set(self.command_config["includes"]))
        merged_config: dict[str, Any] = {}
        for transformed_include in include_transforms:
            cliffy_merger.merge(merged_config, transformed_include.command_config)

        return merged_config

    @classmethod
    def resolve_include_by_path(cls, path) -> Self:
        with open(path, "r") as m:
            return cls(m, as_include=True)

    @staticmethod
    def load_manifest(manifest_io: TextIO) -> dict[str, Any]:
        try:
            manifest_path = os.path.realpath(manifest_io.name)
            all_vars = yaml.safe_load(open(manifest_path, "r")).get("vars", {})
            all_vars["env"] = os.environ
            var_env = Environment(loader=BaseLoader())
            interpolated_vars = {
                var_env.from_string(str(k)).render(all_vars): var_env.from_string(str(v)).render(all_vars)
                for k, v in all_vars.items()
            }
            manifest_env = Environment(loader=FileSystemLoader(manifest_path)).get_template("")
            return yaml.safe_load(manifest_env.render(interpolated_vars))
        except yaml.YAMLError as e:
            out(f"{e}")
            exit_err("~ error loading manifest")



import operator
import os
import platform
import subprocess
import sys
from datetime import datetime
from importlib.resources import files
from pathlib import Path
from typing import Any, NoReturn, Optional

from click import secho
from packaging import version
from pydantic import BaseModel

try:
    from rich.console import Console  # type: ignore
    from rich.table import Table  # type: ignore
except ImportError:
    from .rich import Console, Table

HOME_PATH = str(Path.home())
PYTHON_BIN = f"{sys.exec_prefix}/Scripts" if platform.system() == "Windows" else f"{sys.exec_prefix}/bin"
PYTHON_EXECUTABLE = sys.executable
CLIFFY_CLI_DIR = files("cliffy.clis")
CLIFFY_HOME_PATH = f"{HOME_PATH}/.cliffy"
OPERATOR_MAP = {
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    "<=": operator.le,
    "<": operator.lt,
    ">": operator.gt,
}


class RequirementSpec(BaseModel):
    name: str
    operator: Optional[str]
    version: Optional[str]


def write_to_file(path: str, text: str, executable: bool = False) -> None:
    output_file = Path(path)
    output_file.parent.mkdir(exist_ok=True, parents=True)
    output_file.write_text(text)

    if executable:
        make_executable(path)


def make_executable(path: str) -> None:
    mode = os.stat(path).st_mode
    mode |= (mode & 0o444) >> 2
    os.chmod(path, mode)


def indent_block(block: str, spaces=4) -> str:
    blocklines = block.splitlines()
    return "\n".join([" " * spaces + line for line in blocklines])


def wrap_as_comment(text: str, split_on: Optional[str] = None) -> str:
    if split_on:
        joiner = "\n# "
        return f"# {joiner.join(text.split(split_on))}"

    return f"# {text}"


def wrap_as_var(text: str) -> str:
    return "{{" + text + "}}"


def print_rich_table(cols: list[str], rows: list[list[str]], styles: list[str]) -> None:
    table = Table()
    for style, col in zip(styles, cols):
        table.add_column(col, style=style)
    for row in rows:
        table.add_row(*row)

    console = Console()
    console.print(table)


def get_installed_pip_packages() -> dict[str, str]:
    reqs = subprocess.check_output([sys.executable, "-m", "pip", "freeze"])
    installed_packages = {}
    for r in reqs.split():
        r_spec = r.decode().split("==")
        if len(r_spec) > 1:
            installed_packages[r_spec[0]] = r_spec[1]
    return installed_packages


def parse_requirement(requirement: str) -> RequirementSpec:
    for op in OPERATOR_MAP:
        if op in requirement:
            parts = requirement.replace(" ", "").split(op)
            return RequirementSpec(name=parts[0], operator=op, version=parts[1])

    return RequirementSpec(name=requirement.strip(), operator=None, version=None)


def compare_versions(version1: str, version2: str, op: str) -> bool:
    v1 = version.parse(version1)
    v2 = version.parse(version2)
    return OPERATOR_MAP[op](v1, v2)


def out(text: str, **echo_kwargs: Any) -> None:
    secho(text, **echo_kwargs)


def out_err(text: str) -> None:
    secho(f"{text} 💔", fg="red", err=True)


def exit_err(text: str) -> NoReturn:
    secho(f"{text} 💔", fg="red", err=True)
    raise SystemExit(1)


def age_datetime(date: datetime) -> str:
    delta = datetime.now() - date
    if delta.seconds > 86400:
        return f"{delta.days}d"
    elif delta.seconds > 3600:
        return f"{delta.seconds // 3600}h"
    elif delta.seconds > 60:
        return f"{delta.seconds // 60}m"
    return f"{delta.seconds}s"



from datetime import datetime
from typing import Any, Literal, Union

from pydantic import BaseModel, Field

from ..helper import wrap_as_comment, wrap_as_var

COMMAND_BLOCK = Union[str, list[Union[str, dict[Literal["help"], str]]]]


class CLIManifest(BaseModel):
    name: str = Field(
        description="The name of the CLI. "
        "This will be used as the script name when invoking the CLI from the command line."
    )
    version: str = Field(
        description="The version of the CLI. "
        "This should follow the standard semantic versioning format (e.g., 'MAJOR.MINOR.PATCH')."
    )
    help: str = Field(
        "",
        description="A brief description of the CLI that is displayed when the user invokes the --help or -h option.",
    )
    requires: list[str] = Field(
        [],
        description="List of Python dependencies required for the CLI. Validated on CLI load, update and build. "
        "Supports basic requirements specifier syntax.",
    )
    includes: list[str] = Field(
        [],
        description="List of external CLI manifest paths to include into the main manifest. "
        "Performs a deep merge of manifests sequentially in the order given to assemble a merged manifest. "
        "and finally, deep merges the merged manifest with the main manifest.",
    )
    vars: dict[str, Union[str, dict[str, None]]] = Field(
        {},
        description="A mapping defining manifest variables that can be referenced in any other blocks. "
        "Environments variables can be used in this section with ${some_env_var} for dynamic parsing. "
        "Supports jinja2 formatted expressions as values. "
        "Interpolate defined vars in other blocks jinja2-styled {{ var_name }}.",
    )
    commands: dict[str, COMMAND_BLOCK] = Field(
        {},
        description="A mapping containing the command definitions for the CLI. "
        "Each command should have a unique key- which can be either a group command or nested subcommands. "
        "Nested subcommands are joined by '.' in between each level. "
        "A special (*) wildcard can be used to spread the subcommand to all group-level commands. "
        "The value is the python code to run when the command is called "
        "OR a list of bash commands to run (prefixed with $).",
    )
    imports: Union[str, list[str]] = Field(
        "",
        description="A string block or list of strings containing any module imports. "
        "These can be used to import any python modules that the CLI depends on.",
    )
    functions: list[str] = Field(
        [],
        description="A list containing any helper functions. "
        "Each element of the list can be a separate function. "
        "These functions should be defined as strings that can be executed by the Python interpreter.",
    )
    args: dict[str, list[dict[str, str]]] = Field(
        {},
        description="A mapping containing the arguments and options for each command. "
        "Each key in the mapping should correspond to a command in the commands section. "
        "The value should be a list of mappings representing the params and options for that command.",
    )
    types: dict[str, str] = Field(
        {},
        description="A mapping containing any shared type definitions. "
        "These types can be referenced by name in the args section to provide type annotations "
        "for params and options defined in the args section.",
    )
    cli_options: dict[str, Any] = Field(
        {},
        description="A mapping for any additional options that can be used to customize the behavior of the CLI.",
    )

    @classmethod
    def get_field_description(cls, field_name: str, as_comment: bool = True) -> str:
        field = cls.model_fields.get(field_name)
        if field and field.description:
            if as_comment:
                return wrap_as_comment(field.description, split_on=". ")
            return field.description
        return ""

    @classmethod
    def get_template(cls, name: str) -> str:
        return f"""# cliffy v1 template
manifestVersion: v1

{cls.get_field_description('name')}
name: {name} 

{cls.get_field_description('version')}
version: 0.1.0

{cls.get_field_description('help')}
help: hello world

{cls.get_field_description('includes')}
includes: []

{cls.get_field_description('requires')}
requires: []

{cls.get_field_description('vars')}
vars:
    default_mood: happy
    debug_mode: "{{ env['DEBUG'] or 'False' }}"

{cls.get_field_description('imports')}
imports:
    - import os
    - |
        from collections import defaultdict
        import re

{cls.get_field_description('functions')}
functions:
    - |
        def greet_name(name: str):
            print("hello " + name)

{cls.get_field_description('types')}
types:
    Language: str = typer.Option("english", "-l", help="Language to greet in", prompt=True)

{cls.get_field_description('args')}
args:
    world: [--name|-n: str!]                      # a REQUIRED option
    greet.all: 
        - names: str!                             # a REQUIRED param as denoted by the ! at the end
        - mood: str = "{wrap_as_var("default_mood")}"          # an OPTIONAL param that uses a manifest var as default
        - --language: Language                    # an option with a default that uses Language type as arg definition

{cls.get_field_description('commands')}
commands:
    # this is a parent command that will get invoked with: hello world
    world: 
        - |
            \"\"\"
            Help text for list
            \"\"\"
            greet_name("world")
        - $ echo "i can also mix-and-match this command script to run bash commands"
    
    # this is a nested command that will get invoked with: hello greet all
    greet.all: 
        - help: Help text for list.all       # you can also define help text like this
        - $ echo "hello all"                 # this is a bash command that will get converted to python subprocess call
        - print("greetings from python")     # this python code will get directly invoked

"""

    @classmethod
    def get_raw_template(cls, name: str) -> str:
        return f"""
# cliffy v1 raw template

manifestVersion: v1

{cls.get_field_description('name')}
name: {name} 

{cls.get_field_description('version')}
version: 0.1.0

{cls.get_field_description('help')}
help: 

{cls.get_field_description('includes')}
includes: []

{cls.get_field_description('vars')}
vars: {{}}

{cls.get_field_description('imports')}
imports: []

{cls.get_field_description('functions')}
functions: []

{cls.get_field_description('types')}
types: {{}}

{cls.get_field_description('args')}
args: {{}}

{cls.get_field_description('commands')}
commands: {{}}

"""


class IncludeManifest(BaseModel):
    """Special manifest specifically to define the allowed named objects that can be included"""

    commands: dict[str, Union[str, list[Union[str, dict[Literal["help"], str]]]]] = {}
    imports: Union[str, list[str]] = []
    functions: list[str] = []
    args: dict[str, list[dict[str, str]]] = {}
    types: dict[str, str] = {}
    cli_options: dict[str, str] = {}
    requires: list[str]


class CLIMetadata(BaseModel):
    """Metadata model"""

    cli_name: str
    runner_path: str
    version: str
    loaded: datetime
    manifest: str
    requires: list[str]



import contextlib
import os

from .commander import CLI
from .helper import CLIFFY_CLI_DIR, PYTHON_BIN, PYTHON_EXECUTABLE, write_to_file


class Loader:
    __slots__ = ("cli",)

    def __init__(self, cli: CLI) -> None:
        self.cli = cli

    def deploy_cli(self) -> str:
        cli_path = Loader.get_cli_path(self.cli.name)
        write_to_file(cli_path, self.cli.code)
        return cli_path

    def deploy_script(self) -> str:
        script_path = Loader.get_cli_script_path(self.cli.name)
        write_to_file(script_path, Loader.get_cli_script(self.cli.name), executable=True)
        return script_path

    @classmethod
    def load_from_cli(cls, cli: CLI) -> None:
        L = cls(cli)
        L.deploy_script()
        L.deploy_cli()

    @classmethod
    def unload_cli(cls, cli_name) -> None:
        with contextlib.suppress(FileNotFoundError):
            os.remove(cls.get_cli_script_path(cli_name))
            os.remove(cls.get_cli_path(cli_name))

    @staticmethod
    def get_cli_path(cli_name) -> str:
        return f"{CLIFFY_CLI_DIR}/{cli_name}.py"

    @staticmethod
    def get_cli_script_path(cli_name) -> str:
        return f"{PYTHON_BIN}/{cli_name}"

    @staticmethod
    def get_cli_script(cli_name) -> str:
        return f"""#!{PYTHON_EXECUTABLE}
import sys
from cliffy.clis.{cli_name} import cli

if __name__ == '__main__':
    sys.exit(cli())"""



import os
import sys
from importlib import import_module
from pathlib import Path
from shutil import copy
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Optional

from click.testing import CliRunner, Result
from shiv import cli as shiv_cli
from shiv import pip


def build_cli(
    cli_name: str, script_path: str, deps: Optional[list[str]] = None, output_dir: Optional[str] = None
) -> Result:
    if deps is None:
        deps = []

    if output_dir and not Path(output_dir).exists():
        os.mkdir(output_dir)

    with TemporaryDirectory() as tdist:
        copy(script_path, os.path.join(tdist, f"{cli_name}.py"))
        pip_deps = ["typer"] + deps

        pip.install(["--target", tdist] + pip_deps)

        runner = CliRunner()
        output_file = os.path.join(output_dir, f"{cli_name}") if output_dir else cli_name

        return runner.invoke(
            shiv_cli.main,
            [
                "--site-packages",
                tdist,
                "--compressed",
                "-e",
                f"{cli_name}.cli",
                "-o",
                output_file,
            ],
        )


def run_cli(cli_name: str, script_code: str, args: tuple) -> None:
    with NamedTemporaryFile(mode="w", prefix=f"{cli_name}_", suffix=".py", delete=True) as runner_file:
        runner_file.write(script_code)
        runner_file.flush()
        module_path, module_filename = os.path.split(runner_file.name)
        sys.path.append(module_path)
        module = import_module(module_filename[:-3])

    runner_argvs = [runner_file.name] + list(args)
    sys.argv = runner_argvs
    module.cli()



## Command parser
from typing import Any, Literal, Optional, Tuple, Union

from pybash.transformer import transform as transform_bash

from .manifests import Manifest


class Parser:

    __slots__ = ("manifest",)

    def __init__(self, manifest: Manifest) -> None:
        self.manifest = manifest

    def is_param_required(self, param_type: str) -> bool:
        return (
            param_type.strip().endswith("!")
            if "=" not in param_type
            else param_type.split("=")[0].strip().endswith("!")
        )

    def is_param_option(self, param_name: str) -> bool:
        return "-" in param_name

    def get_default_param_val(self, param_type: str) -> str:
        return param_type.split("=")[1].strip() if "=" in param_type else ""

    def capture_param_aliases(self, param_name: str) -> Tuple[str, list[str]]:
        if "|" not in param_name:
            return param_name.replace("-", ""), []

        base_param_name = param_name.split("|")[0].replace("-", "").strip()
        aliases = param_name.split("|")[1:]

        return base_param_name, aliases

    def parse_command_block(self, script: str) -> str:
        script = transform_bash(script)
        return "".join(" " * 4 + line + "\n" for line in script.split("\n"))

    def parse_command(self, block: Union[str, list[Union[str, dict[Literal["help"], str]]]]) -> str:
        if isinstance(block, list):
            script_block = []
            help_text = ""
            for block_elem in block:
                if isinstance(block_elem, dict):
                    help_text = block_elem.get("help", "")
                else:
                    script_block.append(block_elem)

            code = f'    """\n    {help_text}\n    """\n' if help_text else ""
            code += "".join(map(self.parse_command_block, script_block))
        else:
            code = self.parse_command_block(block)

        return code

    def build_param_type(
        self,
        arg_name: str,
        arg_type: str,
        typer_cls: str,
        aliases: Optional[list[str]] = None,
        default_val: Any = None,
        is_required: bool = False,
    ) -> str:
        parsed_arg_type = f"{arg_name}: {arg_type} = typer.{typer_cls}"

        if not default_val:
            # Required param needs ...
            parsed_arg_type += "(..." if is_required else "(None"
        else:
            parsed_arg_type += f"({default_val.strip()}"

        if aliases:
            parsed_arg_type += f', "--{arg_name}"'
            for alias in aliases:
                parsed_arg_type += f', "{alias.strip()}"'

        parsed_arg_type += "),"
        return parsed_arg_type

    def parse_arg(self, arg_name: str, arg_type: str) -> str:
        is_required = self.is_param_required(arg_type)
        default_val = self.get_default_param_val(arg_type)
        param_type = "Option" if self.is_param_option(arg_name) else "Argument"
        arg_aliases: list[str] = []

        # extract default val before parsing it
        if "=" in arg_type:
            arg_type = arg_type.split("=")[0].strip()

        # strip - before parsing it
        if self.is_param_option(arg_name):
            arg_name, arg_aliases = self.capture_param_aliases(arg_name)

        # strip ! before parsing it
        if is_required:
            arg_type = arg_type[:-1]

        # check for a type def that matches arg_type
        if arg_type in self.manifest.types:
            return f"{arg_name}: {self.manifest.types[arg_type]},"

        return self.build_param_type(
            arg_name,
            arg_type,
            typer_cls=param_type,
            aliases=arg_aliases,
            default_val=default_val,
            is_required=is_required,
        )

    def parse_args(self, command) -> str:
        if not self.manifest.args:
            return ""

        command_args = self.manifest.args.get(command.name)
        if not command_args:
            return ""

        parsed_command_args = ""
        for arg in command_args:
            arg_name, arg_type = next(iter(arg.items()))
            parsed_command_args += f"{self.parse_arg(arg_name.strip(), arg_type.strip())} "

        # strip the extra ", "
        return parsed_command_args[:-2]

    def get_command_func_name(self, command) -> str:
        """a.b -> a_b, c -> c"""
        return command.name.replace(".", "_")

    def get_parsed_command_name(self, command) -> str:
        """a.b -> b or a -> a"""
        return command.name.split(".")[-1] if "." in command.name else command.name

    def to_args(self, d: dict) -> str:
        s = "".join(f" {k}={v}," for k, v in d.items())
        return s[:-1]
